package POWER::GPG;

use version; $VERSION = qv('0.1.0');    # update POD & Changes & README

use warnings;
use strict;
use Carp;

use Perl6::Export::Attrs;
use File::Temp qw( tempfile );      # for verifying detached signatures only :(

use POWER::Utils        v0.3.0;
use POWER::Utils::IO    v0.3.0;


our $DEBUG = 0;                     # 0 - disabled; 1 - warn(); 2 - diag()


#    --trust-model=always           # workaround trust-db headache
#    --keyid-format 0xlong          # long ID more safe
#    --no-permission-warning        # NOTUSED: permissions MUST be secure!!!
my @GPG_Prog = qw(
    /usr/bin/gpg
    --batch
    --no-tty
    --no-options
    --no-secmem-warning
    --yes
    --output -
    --trust-model=always
    --keyid-format 0xlong
);
my $GPG_homedir = undef;


# gpg_config({
#   homedir => $homedir,
#   pass    => { $gpgid1 => $pass1, $gpgid2 => $pass2 },
# });
sub gpg_config :Export(:DEFAULT) {
    my ($opt_ref) = @_;
    croak 'Usage: gpg_config( \%opt )' if @_ != 1;
    croak '{homedir} required' if !exists $opt_ref->{homedir};
    $GPG_homedir = $opt_ref->{homedir};
    return;
}

# $keys_list = gpg_list();
# $keys_list = {
#   $gpg_fingerprint1   => [
#       { email => "...", name => "...", comment => "...", },
#       ...
#       ],
# };
sub gpg_list :Export(:DEFAULT) {
    croak 'Usage: gpg_list()' if @_;
    my %keys;
    my $gpgid;
    my ($exitcode, $out, $err) = _gpg_run(['--fixed-list-mode', '--with-colons',
        '--list-keys', '--with-fingerprint']);
    die $err if $exitcode;
    for (split /\n/, $out) {
        my @col = split /:/;
        $gpgid = $col[9]        if $col[0] eq 'fpr';
        if ($col[0] eq 'uid') {
            my ($name, $comment, $email) = $col[9] =~ /\A
                (.+?)
                (?: \s+ \(([^)]*)\) )?
                (?: \s+ <([^>]*)>   )?
                \z/xms;
            for ($name, $comment, $email) {
                $_ = q{} if !defined $_;
                s/\\x([0-9a-f]{2})/chr(hex($1))/ge;
            }
            push @{ $keys{$gpgid} }, {
                name    => $name,
                comment => $comment,
                email   => $email,
            };
        }
    }
    return \%keys;
}

# @fingerprints = gpg_import($keys);
sub gpg_import :Export(:DEFAULT) {
    my ($keys, $secret) = @_;
    croak 'Usage: gpg_import( $keys )'              if !(1 <= @_ && @_ <= 2);
    croak '$_[0] undefined'                         if !defined $keys;
    if (!$secret) {
    	my (undef, $out) = eval { _gpg_run(['--with-colons'], $keys) };
        croak 'secret key importing prohibited'
            if defined $out && $out =~ /^(?:sec|ssb):/m;
    }
	my ($exitcode, $out, $err, $status) = _gpg_run(['--import'], $keys);
    die $err if $status !~ / IMPORT_OK /;
    #die $err if $status !~ / IMPORT_RES [1-9]/;    # alternative
    my %id = map { ($_ => 1) } $status =~ / IMPORT_OK \d+ (\S+)/g;
    return keys %id;
}

# $public_key = gpg_export($gpgid);
# WARNING    export  $user_id , .. 
#                .
sub gpg_export :Export(:DEFAULT) {
    my ($user_id) = @_;
    croak 'Usage: gpg_export( $user_id )'       if @_ != 1;
    croak '$_[0] undefined'                     if !defined $user_id;
	my ($exitcode, $out, $err) = _gpg_run(['--armor', '--no-emit-version',
        '--export', $user_id]);
    die $err if !length $out;
    return $out;
}

# gpg_delete($gpgid);
# WARNING          ,
#                  .
#                   
#            ,   ݣ    ?
sub gpg_delete :Export(:DEFAULT) {
    my ($user_id) = @_;
    croak 'Usage: gpg_delete( $user_id )'       if @_ != 1;
    croak '$_[0] undefined'                     if !defined $user_id;
	my ($exitcode, $out, $err, $status) = _gpg_run(['--delete-keys', $user_id]);
    die $err if $exitcode;
    #die $err if $status =~ / DELETE_PROBLEM /;     # alternative
    return;
}

# $encrypted = gpg_encrypt($plain, \@encr_to);
sub gpg_encrypt :Export(:DEFAULT) {
    my ($plain, $encr_to, $sign_by) = @_;
    croak 'Usage: gpg_encrypt( $plain, \@encr_to )' if !(2 <= @_ && @_ <= 3);
    croak '$_[0] undefined'                         if !defined $plain;
    croak '$_[1] list is empty'                     if !@$encr_to;
    my @actions = ('--armor', '--no-emit-version','-e');
    push @actions, '-r', $_ for @$encr_to;
    if ($sign_by) {
	    push @actions, '-s';
    	push @actions, '-u', $_ for @$sign_by;
    }
    my ($exitcode, $out, $err, $status) = _gpg_run(\@actions, $plain);
    die $err if $exitcode;
    #die $stderr if !length $stdout;    # alternative
    return $out;
}

# $encrypted = gpg_encrypt_sign($plain, \@encr_to);
# $encrypted = gpg_encrypt_sign($plain, \@encr_to, \@sign_by);
sub gpg_encrypt_sign :Export(:DEFAULT) {
    my ($plain, $encr_to, $sign_by) = @_;
    croak 'Usage: gpg_encrypt_sign( $plain, \@encr_to [, \@sign_by] )'
        if !(2 <= @_ && @_ <= 3);
    return gpg_encrypt($plain, $encr_to, $sign_by || []);
}

# ($plain, $signed_by) = gpg_decrypt($encrypted);
# $signed_by = { $gpg_fingerprint1 => $true, $gpg_longid2 => $false };
sub gpg_decrypt :Export(:DEFAULT) {
    my ($encrypted) = @_;
    croak 'Usage: gpg_decrypt( $encrypted )'    if @_ != 1;
    croak '$_[0] undefined'                     if !defined $encrypted;
    my ($exitcode, $out, $err, $status) = _gpg_run(['-d'], $encrypted);
    die $err
        if $status !~ / BEGIN_DECRYPTION | VALIDSIG | ERRSIG /
        || $status =~ / DECRYPTION_FAILED | BADSIG /
        ;
    my %signed_by = (
        map({ ($_=>1) } $status =~ / VALIDSIG (\S+)/g),
        map({ ($_=>0) } $status =~ / ERRSIG (\S+)/g)
    );
    return ($out, \%signed_by);
}

# $signature = gpg_detachsign($plain);
# $signature = gpg_detachsign($plain, \@sign_by);
sub gpg_detachsign :Export(:DEFAULT) {
    my ($plain, $sign_by) = @_;
    croak 'Usage: gpg_detachsign( $plain [, \@sign_by] )' if !(1<=@_ && @_<=2);
    croak '$_[0] undefined'                               if !defined $plain;
    $sign_by ||= [];
	my @actions = ('--armor','--no-emit-version','--textmode','--detach-sign');
    push @actions, '-u', $_ for @$sign_by;
    my ($exitcode, $out, $err, $status) = _gpg_run(\@actions, $plain);
    die $err if $exitcode;
    return $out;
}

# $signed_by = gpg_verify($plain, $signature);
# $signed_by = { $gpg_fingerprint1 => $true, $gpg_longid2 => $false };
sub gpg_verify :Export(:DEFAULT) {
    my ($plain, $signature) = @_;
    croak 'Usage: gpg_verify( $plain, $signature )' if @_ != 2;
    croak '$_[0] undefined'                         if !defined $plain;
    croak '$_[1] undefined'                         if !defined $signature;
    #         
    #    LD_PRELOAD     (pipe) fd,
    #   detached signature  clearsign. :(
    my ($fh,  $name)  = tempfile();
    print {$fh} $plain;
    close $fh or die "close: $!";
    my ($exitcode, $out, $err, $status) = eval {    # eval for unlink($name)
        _gpg_run(['--verify','-',$name], $signature)
        };
    unlink $name;
    die $@ if $@;
    die $err if $status =~ / BADSIG /
             || $status !~ / VALIDSIG | ERRSIG /
             ;
    my %signed_by = (
        map({ ($_=>1) } $status =~ / VALIDSIG (\S+)/g),
        map({ ($_=>0) } $status =~ / ERRSIG (\S+)/g)
    );
    return \%signed_by;
}

# ($exitcode, $stdout, $stderr, $status) = _gpg_run(\@gpg_options);
# ($exitcode, $stdout, $stderr, $status) = _gpg_run(\@gpg_options, $stdin);
sub _gpg_run {
    my ($actions, $stdin) = @_;
    croak 'you must initialize {homedir} using gpg_config() first!'
        if !defined $GPG_homedir || !length $GPG_homedir;
    # prepare filehandles
    my (%cmd_fh, %r_fh, %w_fh, %buf);
    pipe $r_fh{$_}, $cmd_fh{$_} or die "pipe: $!" for qw(out err status);
    pipe $cmd_fh{$_}, $w_fh{$_} or die "pipe: $!" for qw(in);
    Nonblock($_) for values %r_fh, values %w_fh;
    # initialize buffers
    $buf{$_} = q{} for keys %r_fh;
    $buf{in} = $stdin;
    delete $w_fh{$_}
        for grep { !defined $buf{$_} || !length $buf{$_} } keys %w_fh;
    # run gpg
    local $ENV{LC_ALL} = 'C';
    my @cmd = (@GPG_Prog, "--homedir=$GPG_homedir", '--status-fd=3', @$actions);
    my ($exitcode, $signal) = System({
        cmd     => \@cmd,
        timeout => 30,
        fh      => [@cmd_fh{'in','out','err','status'}],
        cb      => sub {
            close($_) for values %cmd_fh;
            while (keys(%r_fh) + keys(%w_fh)) {
                my ($r, $w) = Select(\%r_fh, \%w_fh);
                for (@$r) {
                    my $is_eof = Sysread($r_fh{$_}, $buf{$_});
                    delete $r_fh{$_} if $is_eof;
                }
                for (@$w) {
                    my $done = eval { Syswrite($w_fh{$_}, $buf{$_}) };
                    die $@ if $@ && $@ !~ /Broken pipe/;
                    delete $w_fh{$_} if $@ || $done;
                }
            }
        },
    });
    _gpg_debug($exitcode, $signal, @buf{'out','err','status'}) if $DEBUG;
    die $buf{err} if $buf{err} =~ /unsafe permissions on homedir/;
    die "gpg killed by signal $signal\n" if $signal;
    return ($exitcode, $buf{out}, $buf{err}, $buf{status});
}

my $NORM    = "\e[0m";
my $BOLD    = "\e[1m";
my $RED     = "\e[31m"; # killed
my $GREEN   = "\e[32m"; # *
my $YELLOW  = "\e[33m"; # exitcode>0
my $CYAN    = "\e[36m"; # known substrings
my $WHITE   = "\e[37m"; # exitcode=0
my $STAR    = "$BOLD$GREEN*$NORM";
sub _gpg_debug {
    my ($exitcode, $signal, $out, $err, $status) = @_;
    $err    =~ s/(unsafe permissions on homedir)/$CYAN$1$NORM/g;
    $status =~ s/(IMPORT_OK|BEGIN_DECRYPTION|DECRYPTION_FAILED|VALIDSIG|ERRSIG|BADSIG)/$CYAN$1$NORM/g;
    my $msg = sprintf " %s %s: %s\n", $STAR, 'EXITCODE',
          $signal   ? "$BOLD$RED(killed by $signal)$NORM"
        : $exitcode ? "$BOLD$YELLOW$exitcode$NORM"
        :             "$BOLD$WHITE$exitcode$NORM"
        ;
    $msg .= sprintf " %s %s:\n%s\n", $STAR, 'STATUS', $status;
    $msg .= sprintf " %s %s:\n%s\n", $STAR, 'STDOUT', $out;
    $msg .= sprintf " %s %s:\n%s\n", $STAR, 'STDERR', $err;
    $msg =~ s/^/# /gm if $DEBUG==2; # simulate diag() for tests
    warn $msg;
}


1; # Magic true value required at end of module
__END__

=head1 NAME

POWER::GPG - Reliable and secure interface to GnuPG.


=head1 VERSION

This document describes POWER::GPG version 0.1.0


=head1 SYNOPSIS

    use POWER::GPG;

    gpg_config({ homedir => '/home/user/.gnupg', });

    $all_keys_list = gpg_list();

    @fingerprints = gpg_import($keys);
    $public_key = gpg_export($gpgid);
    gpg_delete($gpgid);

    $encrypted = gpg_encrypt($plain, \@encr_to);
    $encrypted = gpg_encrypt_sign($plain, \@encr_to);
    $encrypted = gpg_encrypt_sign($plain, \@encr_to, \@sign_by);
    ($plain, $signed_by) = gpg_decrypt($encrypted);

    $signature = gpg_detachsign($plain);
    $signature = gpg_detachsign($plain, \@sign_by);
    $signed_by = gpg_verify($plain, $signature);


=head1 DESCRIPTION

This module isn't designed as comprehensive interface for gpg.
It's goal is provide interface to USE gpg, not to MANAGE it.
This mean you must manually initialize gpg homedir; manage secret keys, etc.
and then use this module to do simple public key management and
encrypt/decrypt/sign/verify operations.

CPAN already contains a lot of modules for accessing GnuPG.
This one was developed because all these modules not reliable
enough and some of them not secure. :(

This module execute gpg with correct signal handling; use non-blocking
pipes and multiplexing to avoid hang on processing huge files;
never store sensitive information on hard drive in temporary files
(one temporary file used for checking detached signature); use timeout
to protect against gpg hands.

Limitations of other modules from reliability/security view:

=over

=item GPG

 - Doesn't handle SIGCHLD, SIGINT, SIGQUIT.
 - Hang on large files.
 - Parse unreliable STDOUT instead of reliable --status-fd.

=item Crypt::GPG

 - Doesn't handle SIGINT, SIGQUIT.
 - Incorrectly handle SIGCHLD.

=item GnuPG

 - Doesn't handle SIGCHLD, SIGINT, SIGQUIT.
 - Use deprecated shared memory interface.
 - Use temporary files to store sensitive information.

=item GnuPG::Interface

 - Doesn't handle SIGCHLD, SIGINT, SIGQUIT.

=back


=head1 INTERFACE

All these functions will throw exception on error. List of possible
exceptions and their reasons see in L</DIAGNOSTICS> below.

General rule is: if function can't do requested operation (import key or
decrypt message, for example) it throw exception. B<Incorrect signature
also result in exception>.

Some specific cases result in exception documented below, but you shouldn't
read this as "only possible exceptions for this function".

=over

=item gpg_config( \%opt )

Configure gpg options. Usually it's enough to call gpg_config() once when
your program initialize and before it begin using other gpg_*() functions,

At this time %opt can have only one key {homedir} which should point to
existing GnuPG homedir. In the future %opt will be also used to configure
passwords for secret keys.

Return: nothing.


=item gpg_list( )

Return: list of all public keys as HASHREF. Keys is this hash are key's
fingerprints. Values - ref to array of hashes for each 'UID' of this key.
Hashes with UID's have three keys: {email}, {name}, {comment}.

 $keys_list = {
    $gpg_fingerprint1   => [
        { email => "...", name => "...", comment => "..." },
        ...
        ],
    ...
 };


=item gpg_import( $keys )

Import one or more public or secret keys from $keys into GnuPG.
Only public keys importing allowed.

Throw exception on importing secret keys. 

Return: list of fingerprints for imported keys.


=item gpg_export( $user_id )

Return: public key(s) matching $user_id, ascii-armored.


=item gpg_delete( $user_id )

Delete public key for $user_id.

Throw exception if secret key for $user_id also exists.

Return: nothing.


=item gpg_encrypt( $plain, \@encr_to )

Encrypt message in $plain to recipients in @encr_to.

Throw exception is even single recipient from @encr_to not found not found
in public keyring.

Return: encrypted message, ascii-armored.


=item gpg_encrypt_sign( $plain, \@encr_to [, \@sign_by] )

Like gpg_encrypt(), but also sign message. If @sign_by not given or [],
then sign by default secret key.


=item gpg_decrypt( $encrypted )

Decrypt/check signature. $encrypted can be 1) encrypted; 2) encrypted
and signed; 3) clearsigned.

Return: list with two elements ($plain, $sign_by). $plain is decrypted
(or just cut from $encrypted in clearsigned case) message. $sign_by is
HASHREF with information about signatures, if $encrypted was signed.
This hash keys will be key "fingerprint" (with 1 as value) for all
verified signatures and gpg "longid" (with 0 as value) for all
signatures which was not verified because their public keys isn't
imported into our keyring.

 $signed_by = {
    $gpg_fingerprint1   => 1,
    $gpg_longid2        => 0,
    ...
 };


=item gpg_detachsign( $plain [, \@sign_by] )

Create detached signature for $plain. If @sign_by not given or [],
then sign by default secret key.

Return: detached signature, ascii-armored.


=item gpg_verify( $plain, $signature )

Verify detached signature for $plain.

Return: HASHREF with information about signatures (see $sign_by returned
value in gpg_decrypt().


=back

=head1 DIAGNOSTICS

=over

=item C<< Usage: gpg_config( \%opt ) >>

=item C<< Usage: gpg_list() >>

=item C<< Usage: gpg_import( $public_key ) >>

=item C<< Usage: gpg_export( $user_id ) >>

=item C<< Usage: gpg_delete( $user_id ) >>

=item C<< Usage: gpg_encrypt( $plain, \@encr_to ) >>

=item C<< Usage: gpg_encrypt_sign( $plain, \@encr_to [, \@sign_by] ) >>

=item C<< Usage: gpg_decrypt( $encrypted ) >>

=item C<< Usage: gpg_detachsign( $plain [, \@sign_by] ) >>

=item C<< Usage: gpg_verify( $plain, $signature ) >>

You've executed function with incorrect number of params.


=item C<< $_[0] undefined >>

=item C<< $_[1] undefined >>

Most gpg_*() functions require their params be at least defined.


=item C<< $_[1] list is empty >>

gpg_encrypt() or gpg_encrypt_sign() was executed with empty @encr_to param.


=item C<< {homedir} required >>

gpg_config() was executed without {homedir} key in hashref.


=item C<< you must initialize {homedir} using gpg_config() first! >>

You tried to execute one of gpg_*() functions before gpg_config().


=item C<< secret key importing prohibited >>

You tried to import secret key.


=item C<< gpg killed by signal %d >>

No idea why this may happen. :)


=item C<< ... GPG STDERR ... >>

 - Unsafe permissions on homedir. In gpg this is a warning,
   but in this module it turned into fatal exception.
 - gpg initialization error.
 - Can't import key.
 - Can't export key.
 - Can't delete key.
 - Can't encrypt.
 - Can't decrypt.
 - Can't sign.
 - Bad signature.


=item C<< ... ANY OTHER I/O-RELATED ERROR ... >>

pipe/sysread/syswrite/select/close/...etc... operations failed.


=back


=head1 CONFIGURATION AND ENVIRONMENT

POWER::GPG requires no configuration files or environment variables.


=head1 DEPENDENCIES

 Perl6::Export::Attrs

 POWER::Utils	    v0.3.0
 POWER::Utils::IO   v0.3.0


=head1 INCOMPATIBILITIES

None reported.


=head1 BUGS AND LIMITATIONS

GnuPG has too many features to scrupulously test them all,
so some things doesn't tested with this module. :(

Limitations:

 IN TODO:
 - No support for password-protected secret keys.
 - No support for checking key signatures.
 NOT IN TODO:
 - No support for secret keys list/export/delete.
 - No way to detect possible problem with {homedir} in
   gpg_verify() or gpg_decrypt($clearsigned).
 - No support for trustdb.

No bugs have been reported.

Please report any bugs or feature requests to
C<bug-power-gpg@rt.cpan.org>, or through the web interface at
L<http://rt.cpan.org>.


=head1 AUTHOR

Alex Efros  <powerman@asdfgroup.com>


=head1 LICENCE AND COPYRIGHT

Copyright (c) 2006, Alex Efros <powerman@asdfgroup.com>. All rights reserved.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.


=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
